@@ -14,6 +14,7 @@ from crytic_compile import cryticparser
 from crytic_compile import CryticCompile, InvalidCompilation
 
 from .utils import color_good, color_warn
+from .byte2op import mnemonic
 from .sevm import *
 
 if hasattr(sys, 'set_int_max_str_digits'): # Python verion >=3.8.14, >=3.9.14, >=3.10.7, or >=3.11
@@ -45,6 +46,8 @@ def parse_args(args) -> argparse.Namespace:
 
     parser.add_argument('--solver-timeout-branching', metavar='TIMEOUT', type=int, default=1000, help='set timeout (in milliseconds) for solving branching conditions (default: %(default)s)')
     parser.add_argument('--solver-timeout-assertion', metavar='TIMEOUT', type=int, default=60000, help='set timeout (in milliseconds) for solving assertion violation conditions (default: %(default)s)')
+    parser.add_argument('--solver-fresh', action='store_true', help='run an extra solver with a fresh state for unknown')
+    parser.add_argument('--solver-axioms', action='store_true', help='run an extra solver with axioms for unknown')
     parser.add_argument('--solver-subprocess', action='store_true', help='run an extra solver in subprocess for unknown')
 
     parser.add_argument('-v', '--verbose', action='count', default=0, help='increase verbosity levels: -v, -vv, -vvv, -vvvv')
@@ -145,7 +148,20 @@ def mk_callvalue() -> Word:
     return BitVec('msg_value', 256)
 
 def mk_balance() -> Word:
-    return BitVec('this_balance', 256)
+    return Array('balance0', BitVecSort(256), BitVecSort(256))
+
+def mk_block() -> Block:
+    block = Block(
+        basefee    = ZeroExt(160, BitVec('block_basefee', 96)),     # practical limit 96 bit
+        chainid    = ZeroExt(192, BitVec('block_chainid', 64)),     # chainid 64 bit
+        coinbase   = ZeroExt(96, BitVec('block_coinbase', 160)),    # address 160 bit
+        difficulty = BitVec('block_difficulty', 256),
+        gaslimit   = ZeroExt(160, BitVec('block_gaslimit', 96)),    # practical limit 96 bit
+        number     = ZeroExt(192, BitVec('block_number', 64)),      # practical limit 64 bit
+        timestamp  = ZeroExt(192, BitVec('block_timestamp', 64)),   # practical limit 64 bit
+    )
+    block.chainid = con(1) # for ethereum
+    return block
 
 def mk_caller(solver) -> Word:
     caller = BitVec('msg_sender', 256)
@@ -170,6 +186,7 @@ def run_bytecode(hexcode: str, args: argparse.Namespace, options: Dict) -> List[
     solver = mk_solver(args)
 
     balance = mk_balance()
+    block = mk_block()
     callvalue = mk_callvalue()
     caller = mk_caller(solver)
     this = mk_this(solver)
@@ -179,7 +196,8 @@ def run_bytecode(hexcode: str, args: argparse.Namespace, options: Dict) -> List[
         pgm       = { this: pgm },
         code      = { this: code },
         storage   = { this: storage },
-        balance   = { this: balance },
+        balance   = balance,
+        block     = block,
         calldata  = [],
         callvalue = callvalue,
         caller    = caller,
@@ -192,11 +210,11 @@ def run_bytecode(hexcode: str, args: argparse.Namespace, options: Dict) -> List[
     models = []
     for idx, ex in enumerate(exs):
         opcode = ex.pgm[ex.this][ex.pc].op[0]
-        if is_bv_value(opcode) and opcode.as_long() in [EVM.STOP, EVM.RETURN, EVM.REVERT]:
+        if is_bv_value(opcode) and opcode.as_long() in [EVM.STOP, EVM.RETURN, EVM.REVERT, EVM.INVALID]:
             gen_model(args, models, idx, ex)
-            print(f'Final opcode: {opcode.as_long()} | Return data: {ex.output} | Input example: {models[-1][0]}')
+            print(f'Final opcode: {mnemonic(opcode)} | Return data: {ex.output} | Input example: {models[-1][0]}')
         else:
-            print(color_warn('Not supported: ' + opcode + ' ' + ex.error))
+            print(color_warn(f'Not supported: {mnemonic(opcode)} {ex.error}'))
         if args.verbose >= 1:
             print(f'# {idx+1} / {len(exs)}')
             print(ex)
@@ -225,7 +243,8 @@ def setup(
         pgm       = { this: pgm },
         code      = { this: code },
         storage   = { this: {} },
-        balance   = { this: con(0) },
+        balance   = mk_balance(),
+        block     = mk_block(),
         calldata  = [],
         callvalue = con(0),
         caller    = mk_caller(solver),
@@ -265,6 +284,8 @@ def run(
     args: argparse.Namespace,
     options: Dict
 ) -> int:
+    if args.debug: print(f'Executing {funname}')
+
     #
     # calldata
     #
@@ -282,12 +303,6 @@ def run(
 
     callvalue = mk_callvalue()
 
-    #
-    # balance
-    #
-
-    balance = mk_balance()
-
     #
     # run
     #
@@ -304,7 +319,9 @@ def run(
         pgm       = setup_ex.pgm.copy(), # shallow copy
         code      = setup_ex.code.copy(), # shallow copy
         storage   = deepcopy(setup_ex.storage),
-        balance   = { setup_ex.this: sevm.arith(EVM.ADD, balance, callvalue) },
+        balance   = setup_ex.balance, # TODO: add callvalue
+        #
+        block     = deepcopy(setup_ex.block),
         #
         calldata  = cd,
         callvalue = callvalue,
@@ -316,6 +333,7 @@ def run(
         jumpis    = {},
         output    = None,
         symbolic  = True,
+        prank     = Prank(), # prank is reset after setUp()
         #
         solver    = solver,
         path      = deepcopy(setup_ex.path),
@@ -324,6 +342,7 @@ def run(
         cnts      = deepcopy(setup_ex.cnts),
         sha3s     = deepcopy(setup_ex.sha3s),
         storages  = deepcopy(setup_ex.storages),
+        balances  = deepcopy(setup_ex.balances),
         calls     = deepcopy(setup_ex.calls),
         failed    = setup_ex.failed,
         error     = setup_ex.error,
@@ -334,13 +353,15 @@ def run(
     models = []
     stuck = []
     for idx, ex in enumerate(exs):
+        if args.debug: print(f'Checking output: {idx+1} / {len(exs)}')
+
         opcode = ex.pgm[ex.this][ex.pc].op[0]
         if is_bv_value(opcode) and opcode.as_long() in [EVM.STOP, EVM.RETURN]:
             if ex.failed:
                 gen_model(args, models, idx, ex)
             else:
                 normal += 1
-        elif is_bv_value(opcode) and opcode.as_long() == EVM.REVERT:
+        elif is_bv_value(opcode) and opcode.as_long() in [EVM.REVERT, EVM.INVALID]:
             # Panic(1) # bytes4(keccak256("Panic(uint256)")) + bytes32(1)
             if ex.output == int('4e487b71' + '0000000000000000000000000000000000000000000000000000000000000001', 16): # 152078208365357342262005707660225848957176981554335715805457651098985835139029979365377
                 gen_model(args, models, idx, ex)
@@ -359,14 +380,14 @@ def run(
     print(f"{passfail} {funsig} (paths: {normal}/{len(exs)}, time: {end - start:0.2f}s, bounds: [{', '.join(dyn_param_size)}])")
     for model, idx, ex in models:
         if model:
-            print(color_warn('Counterexample: ' + str_model(model, args)))
+            print(color_warn(f'Counterexample: {str_model(model, args)}'))
         else:
-            print(color_warn('Counterexample: unknown'))
+            print(color_warn(f'Counterexample: unknown'))
         if args.verbose >= 1:
             print(f'# {idx+1} / {len(exs)}')
             print(ex)
     for opcode, idx, ex in stuck:
-        print(color_warn('Not supported: ' + opcode + ' ' + ex.error))
+        print(color_warn(f'Not supported: {mnemonic(opcode)} {ex.error}'))
         if args.verbose >= 1:
             print(f'# {idx+1} / {len(exs)}')
             print(ex)
@@ -390,15 +411,21 @@ def run(
         return 1
 
 def gen_model(args: argparse.Namespace, models: List, idx: int, ex: Exec) -> None:
+    if args.debug: print(f'{" "*2}Checking assertion violation')
+
     res = ex.solver.check()
-    if res == sat: model = ex.solver.model()
-    if res == unknown:
+    if res == sat:
+        if args.debug: print(f'{" "*4}Generating a counterexample')
+        model = ex.solver.model()
+    if res == unknown and args.solver_fresh:
+        if args.debug: print(f'{" "*4}Checking again with a fresh solver')
         sol2 = SolverFor('QF_AUFBV', ctx=Context())
         sol2.set(timeout=args.solver_timeout_assertion)
         sol2.from_string(ex.solver.sexpr())
         res = sol2.check()
         if res == sat: model = sol2.model()
-    if res == sat and not is_valid_model(model):
+    if res == sat and not is_valid_model(model) and args.solver_axioms:
+        if args.debug: print(f'{" "*4}Checking again with axioms')
         ctx = Context()
         sol3 = Solver(ctx=ctx)
         sol3.set(timeout=args.solver_timeout_assertion)
@@ -431,6 +458,7 @@ def gen_model(args: argparse.Namespace, models: List, idx: int, ex: Exec) -> Non
         res = sol3.check()
         if res == sat: model = sol3.model()
     if res == unknown and args.solver_subprocess:
+        if args.debug: print(f'{" "*4}Checking again in an external process')
         fname = f'/tmp/{uuid.uuid4().hex}.smt2'
         if args.verbose >= 4: print(f'z3 -smt2 {fname}')
         with open(fname, 'w') as f:
@@ -441,13 +469,17 @@ def gen_model(args: argparse.Namespace, models: List, idx: int, ex: Exec) -> Non
         if res_str == 'unsat':
             res = unsat
     if res == unsat:
+        if args.debug: print(f'{" "*4}Passed')
         return
     if res == sat:
         if is_valid_model(model):
+            if args.debug: print(f'{" "*4}Done')
             models.append((model, idx, ex))
         else:
+            if args.debug: print(f'{" "*4}Invalid counterexample')
             models.append((None, idx, ex))
     else:
+        if args.debug: print(f'{" "*4}Timeout')
         models.append((None, idx, ex))
 
 def is_valid_model(model) -> bool:
