@@ -9,8 +9,9 @@ from typing import List, Dict, Set, Tuple, Any
 from functools import reduce
 
 from z3 import *
-from .byte2op import Opcode, decode, concat
-from .utils import EVM, color_good, color_warn, hevm_cheat_code, sha3_inv
+from .byte2op import Opcode, decode, concat, mnemonic
+from .utils import EVM, color_good, color_warn, sha3_inv
+from .cheatcodes import hevm_cheat_code, Prank
 
 Word = Any # z3 expression (including constants)
 Byte = Any # z3 expression (including constants)
@@ -21,19 +22,12 @@ Steps = Dict[int,Dict[str,Any]] # execution tree
 # symbolic states
 f_calldataload = Function('calldataload', BitVecSort(256), BitVecSort(256)) # index
 f_calldatasize = Function('calldatasize', BitVecSort(256))
-f_origin       = Function('origin'      , BitVecSort(256))
-f_coinbase     = Function('coinbase'    , BitVecSort(256))
 f_extcodesize  = Function('extcodesize' , BitVecSort(256), BitVecSort(256)) # target address
 f_extcodehash  = Function('extcodehash' , BitVecSort(256), BitVecSort(256)) # target address
 f_blockhash    = Function('blockhash'   , BitVecSort(256), BitVecSort(256)) # block number
 f_gas          = Function('gas'         , BitVecSort(256), BitVecSort(256)) # cnt
 f_gasprice     = Function('gasprice'    , BitVecSort(256))
-f_timestamp    = Function('timestamp'   , BitVecSort(256))
-f_blocknumber  = Function('blocknumber' , BitVecSort(256))
-f_difficulty   = Function('difficulty'  , BitVecSort(256))
-f_gaslimit     = Function('gaslimit'    , BitVecSort(256))
-f_chainid      = Function('chainid'     , BitVecSort(256))
-f_orig_balance = Function('orig_balance', BitVecSort(256), BitVecSort(256)) # target address
+f_origin       = Function('origin'      , BitVecSort(256))
 
 # uninterpreted arithmetic
 f_add  = Function('evm_add' , BitVecSort(256), BitVecSort(256), BitVecSort(256))
@@ -164,12 +158,32 @@ class State:
         else:
             return None
 
+class Block:
+    basefee: Any
+    chainid: Any
+    coinbase: Any
+    difficulty: Any # prevrandao
+    gaslimit: Any
+    number: Any
+    timestamp: Any
+
+    def __init__(self, **kwargs) -> None:
+        self.basefee    = kwargs['basefee']
+        self.chainid    = kwargs['chainid']
+        self.coinbase   = kwargs['coinbase']
+        self.difficulty = kwargs['difficulty']
+        self.gaslimit   = kwargs['gaslimit']
+        self.number     = kwargs['number']
+        self.timestamp  = kwargs['timestamp']
+
 class Exec: # an execution path
     # network
     pgm: Dict[Any,List[Opcode]] # address -> { opcode map: pc -> opcode }
     code: Dict[Any,List[Any]] # address -> opcode sequence
     storage: Dict[Any,Dict[int,Any]] # address -> { storage slot -> value }
-    balance: Dict[Any,Any] # address -> balance
+    balance: Any # address -> balance
+    # block
+    block: Block
     # tx
     calldata: List[Byte] # msg.data
     callvalue: Word # msg.value
@@ -181,6 +195,7 @@ class Exec: # an execution path
     jumpis: Dict[str,Dict[bool,int]] # for loop detection
     output: Any # returndata
     symbolic: bool # symbolic or concrete storage
+    prank: Prank
     # path
     solver: Solver
     path: List[Any] # path conditions
@@ -188,7 +203,8 @@ class Exec: # an execution path
     log: List[Tuple[List[Word], Any]] # event logs emitted
     cnts: Dict[int,int] # opcode -> frequency
     sha3s: List[Tuple[Word,Word]] # sha3 hashes generated
-    storages: List[Tuple[Any,Any]] # storage updates
+    storages: Dict[Any,Any] # storage updates
+    balances: Dict[Any,Any] # balance updates
     calls: List[Any] # external calls
     failed: bool
     error: str
@@ -199,6 +215,8 @@ class Exec: # an execution path
         self.storage  = kwargs['storage']
         self.balance  = kwargs['balance']
         #
+        self.block    = kwargs['block']
+        #
         self.calldata = kwargs['calldata']
         self.callvalue= kwargs['callvalue']
         self.caller   = kwargs['caller']
@@ -209,6 +227,7 @@ class Exec: # an execution path
         self.jumpis   = kwargs['jumpis']
         self.output   = kwargs['output']
         self.symbolic = kwargs['symbolic']
+        self.prank    = kwargs['prank']
         #
         self.solver   = kwargs['solver']
         self.path     = kwargs['path']
@@ -217,6 +236,7 @@ class Exec: # an execution path
         self.cnts     = kwargs['cnts']
         self.sha3s    = kwargs['sha3s']
         self.storages = kwargs['storages']
+        self.balances = kwargs['balances']
         self.calls    = kwargs['calls']
         self.failed   = kwargs['failed']
         self.error    = kwargs['error']
@@ -234,15 +254,16 @@ class Exec: # an execution path
         return ''.join([
             'PC: '              , str(self.this), ' ', str(self.pc), ' ', str(self.pgm[self.this][self.pc]), '\n',
             str(self.st),
+            'Balance: '         , str(self.balance), '\n',
             'Storage:\n'        , ''.join(map(lambda x: '- ' + str(x) + ': ' + str(self.storage[x]) + '\n', self.storage)),
-            'Balance:\n'        , ''.join(map(lambda x: '- ' + str(x) + ': ' + str(self.balance[x]) + '\n', self.balance)),
         #   'Solver:\n'         , self.str_solver(), '\n',
             'Path:\n'           , self.str_path(),
             'Output: '          , str(self.output) , '\n',
             'Log: '             , str(self.log)    , '\n',
         #   'Opcodes:\n'        , self.str_cnts(),
         #   'Memsize: '         , str(len(self.st.memory)), '\n',
-            'Storage updates:\n', ''.join(map(lambda x: '- ' + str(x) + '\n', self.storages)),
+            'Balance updates:\n', ''.join(map(lambda x: '- ' + str(x) + '\n', sorted(self.balances.items(), key=lambda x: str(x[0])))),
+            'Storage updates:\n', ''.join(map(lambda x: '- ' + str(x) + '\n', sorted(self.storages.items(), key=lambda x: str(x[0])))),
             'SHA3 hashes:\n'    , ''.join(map(lambda x: '- ' + str(x) + '\n', self.sha3s)),
             'External calls:\n' , ''.join(map(lambda x: '- ' + str(x) + '\n', self.calls)),
         #   'Calldata: '        , str(self.calldata), '\n',
@@ -253,10 +274,39 @@ class Exec: # an execution path
         while self.pgm[self.this][self.pc] is None:
             self.pc += 1
 
+    def check(self, cond: Any) -> Any:
+        self.solver.push()
+        self.solver.add(simplify(cond))
+        result = self.solver.check()
+        self.solver.pop()
+        return result
+
+    def select(self, array: Any, key: Word, arrays: Dict) -> Word:
+        if array in arrays:
+            store = arrays[array]
+            if store.decl().name() == 'store' and store.num_args() == 3:
+                base = store.arg(0)
+                key0 = store.arg(1)
+                val0 = store.arg(2)
+                if eq(key, key0): # structural equality
+                    return val0
+                if self.check(key == key0) == unsat: # key != key0
+                    return self.select(base, key, arrays)
+                if self.check(key != key0) == unsat: # key == key0
+                    return val0
+        return Select(array, key)
+
     def balance_of(self, addr: Word) -> Word:
-        if addr not in self.balance:
-            self.balance[addr] = f_orig_balance(addr)
-        return self.balance[addr]
+        value = self.select(self.balance, addr, self.balances)
+        self.solver.add(ULT(value, con(2**96))) # practical assumption on the max balance per account
+        return value
+
+    def balance_update(self, addr: Word, value: Word):
+        new_balance_var = Array(f'balance{1+len(self.balances)}', BitVecSort(256), BitVecSort(256))
+        new_balance = Store(self.balance, addr, value)
+        self.solver.add(new_balance_var == new_balance)
+        self.balance = new_balance_var
+        self.balances[new_balance_var] = new_balance
 
     def sinit(self, slot: int, keys) -> None:
         if slot not in self.storage[self.this]:
@@ -281,7 +331,7 @@ class Exec: # an execution path
         if len(keys) == 0:
             return self.storage[self.this][slot][0]
         else:
-            return Select(self.storage[self.this][slot][len(keys)], concat(keys))
+            return self.select(self.storage[self.this][slot][len(keys)], concat(keys), self.storages)
 
     def sstore(self, loc: Any, val: Any) -> None:
         offsets = self.decode_storage_loc(loc)
@@ -295,7 +345,7 @@ class Exec: # an execution path
             new_storage = Store(self.storage[self.this][slot][len(keys)], concat(keys), val)
             self.solver.add(new_storage_var == new_storage)
             self.storage[self.this][slot][len(keys)] = new_storage_var
-            self.storages.append((new_storage_var,new_storage))
+            self.storages[new_storage_var] = new_storage
 
     def decode_storage_loc(self, loc: Any) -> Any:
         def normalize(expr: Any) -> Any:
@@ -631,7 +681,11 @@ class SEVM:
         if not arg_size >= 0: raise ValueError(arg_size)
         if not ret_size >= 0: raise ValueError(ret_size)
 
-        ex.balance[ex.this] = self.arith(EVM.SUB, ex.balance_of(ex.this), fund)
+        caller = ex.prank.lookup(ex.this, to)
+
+        if not (is_bv_value(fund) and fund.as_long() == 0):
+            ex.balance_update(caller, self.arith(EVM.SUB, ex.balance_of(caller), fund))
+            ex.balance_update(to,     self.arith(EVM.ADD, ex.balance_of(to),     fund))
 
         def call_known() -> None:
             calldata = [None] * arg_size
@@ -645,9 +699,11 @@ class SEVM:
                 storage   = ex.storage,
                 balance   = ex.balance,
                 #
+                block     = ex.block,
+                #
                 calldata  = calldata,
                 callvalue = fund,
-                caller    = ex.this,
+                caller    = caller,
                 this      = to,
                 #
                 pc        = 0,
@@ -655,6 +711,7 @@ class SEVM:
                 jumpis    = {},
                 output    = None,
                 symbolic  = ex.symbolic,
+                prank     = Prank(),
                 #
                 solver    = ex.solver,
                 path      = ex.path,
@@ -663,6 +720,7 @@ class SEVM:
                 cnts      = ex.cnts,
                 sha3s     = ex.sha3s,
                 storages  = ex.storages,
+                balances  = ex.balances,
                 calls     = ex.calls,
                 failed    = ex.failed,
                 error     = ex.error,
@@ -684,6 +742,7 @@ class SEVM:
                 new_ex.jumpis = deepcopy(ex.jumpis)
                 # new_ex.output is passed into the caller
                 new_ex.symbolic = ex.symbolic
+                new_ex.prank = ex.prank
 
                 # set return data (in memory)
                 wstore_partial(new_ex.st.memory, ret_loc, 0, min(ret_size, new_ex.returndatasize()), new_ex.output, new_ex.returndatasize())
@@ -700,7 +759,7 @@ class SEVM:
                     stack.append((new_ex, step_id))
                 else:
                     # got stuck during external call
-                    new_ex.error = str('external call stuck: ' + opcode)
+                    new_ex.error = f'external call stuck: {mnemonic(opcode)}'
                     out.append(new_ex)
 
         def call_unknown() -> None:
@@ -731,11 +790,12 @@ class SEVM:
                 # vm.fail()
                 if arg == hevm_cheat_code.fail_payload: # BitVecVal(hevm_cheat_code.fail_payload, 800)
                     ex.failed = True
-                # vm.assume()
+                # vm.assume(bool)
                 elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.assume_sig:
                     assume_cond = simplify(is_non_zero(Extract(255, 0, arg)))
                     ex.solver.add(assume_cond)
                     ex.path.append(str(assume_cond))
+                # vm.getCode(string)
                 elif simplify(Extract(arg_size*8-1, arg_size*8-32, arg)) == hevm_cheat_code.get_code_sig:
                     calldata = bytes.fromhex(hex(arg.as_long())[2:])
                     path_len = int.from_bytes(calldata[36:68], 'big')
@@ -766,6 +826,46 @@ class SEVM:
                     ret_bytes = bytes.fromhex(ret_bytes)
 
                     ret = BitVecVal(int.from_bytes(ret_bytes, 'big'), ret_len * 8)
+                # vm.prank(address)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.prank_sig:
+                    result = ex.prank.prank(simplify(Extract(255, 0, arg)))
+                    if not result:
+                        ex.error = 'You have an active prank already.'
+                        out.append(ex)
+                        return
+                # vm.startPrank(address)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.start_prank_sig:
+                    result = ex.prank.startPrank(simplify(Extract(255, 0, arg)))
+                    if not result:
+                        ex.error = 'You have an active prank already.'
+                        out.append(ex)
+                        return
+                # vm.stopPrank()
+                elif eq(arg.sort(), BitVecSort((4)*8)) and simplify(Extract(31, 0, arg)) == hevm_cheat_code.stop_prank_sig:
+                    ex.prank.stopPrank()
+                # vm.deal(address,uint256)
+                elif eq(arg.sort(), BitVecSort((4+32*2)*8)) and simplify(Extract(543, 512, arg)) == hevm_cheat_code.deal_sig:
+                    who = simplify(Extract(511, 256, arg))
+                    amount = simplify(Extract(255, 0, arg))
+                    ex.balance_update(who, amount)
+                # vm.fee(uint256)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.fee_sig:
+                    ex.block.basefee = simplify(Extract(255, 0, arg))
+                # vm.chainId(uint256)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.chainid_sig:
+                    ex.block.chainid = simplify(Extract(255, 0, arg))
+                # vm.coinbase(address)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.coinbase_sig:
+                    ex.block.coinbase = simplify(Extract(255, 0, arg))
+                # vm.difficulty(uint256)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.difficulty_sig:
+                    ex.block.difficulty = simplify(Extract(255, 0, arg))
+                # vm.roll(uint256)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.roll_sig:
+                    ex.block.number = simplify(Extract(255, 0, arg))
+                # vm.warp(uint256)
+                elif eq(arg.sort(), BitVecSort((4+32)*8)) and simplify(Extract(287, 256, arg)) == hevm_cheat_code.warp_sig:
+                    ex.block.timestamp = simplify(Extract(255, 0, arg))
                 else:
                     # TODO: support other cheat codes
                     ex.error = str('Unsupported cheat code: calldata: ' + str(arg))
@@ -802,17 +902,22 @@ class SEVM:
 
         # new account address
         new_addr = create_address(ex.cnt_create())
+        for addr in ex.pgm:
+            ex.solver.add(new_addr != addr) # ensure new address is fresh
 
         # setup new account
         ex.pgm[new_addr] = create_pgm   # existing pgm must be empty
         ex.code[new_addr] = create_code # existing code must be empty
         ex.storage[new_addr] = {}       # existing storage may not be empty and reset here
-        ex.balance[new_addr] = f_orig_balance(new_addr)
+
+        # lookup prank
+        caller = ex.prank.lookup(ex.this, new_addr)
 
         # transfer value
-        ex.solver.add(UGE(ex.balance_of(ex.this), value)) # assume balance is enough; otherwise ignore this path
-        ex.balance[ex.this] = self.arith(EVM.SUB, ex.balance_of(ex.this), value)
-        ex.balance[new_addr] = self.arith(EVM.ADD, ex.balance_of(new_addr), value)
+        ex.solver.add(UGE(ex.balance_of(caller), value)) # assume balance is enough; otherwise ignore this path
+        if not (is_bv_value(value) and value.as_long() == 0):
+            ex.balance_update(caller,   self.arith(EVM.SUB, ex.balance_of(caller),   value))
+            ex.balance_update(new_addr, self.arith(EVM.ADD, ex.balance_of(new_addr), value))
 
         # execute contract creation code
         (new_exs, new_steps) = self.run(Exec(
@@ -821,6 +926,8 @@ class SEVM:
             storage   = ex.storage,
             balance   = ex.balance,
             #
+            block     = ex.block,
+            #
             calldata  = [],
             callvalue = value,
             caller    = ex.this,
@@ -831,6 +938,7 @@ class SEVM:
             jumpis    = {},
             output    = None,
             symbolic  = False,
+            prank     = Prank(),
             #
             solver    = ex.solver,
             path      = ex.path,
@@ -839,6 +947,7 @@ class SEVM:
             cnts      = ex.cnts,
             sha3s     = ex.sha3s,
             storages  = ex.storages,
+            balances  = ex.balances,
             calls     = ex.calls,
             failed    = ex.failed,
             error     = ex.error,
@@ -872,6 +981,7 @@ class SEVM:
                 new_ex.jumpis = deepcopy(ex.jumpis)
                 new_ex.output = None # output is reset, not restored
                 new_ex.symbolic = ex.symbolic
+                new_ex.prank = ex.prank
 
                 # push new address to stack
                 new_ex.st.push(new_addr)
@@ -939,7 +1049,7 @@ class SEVM:
                 ex.solver.add(target_reachable)
                 if ex.solver.check() != unsat: # jump
                     if self.options.get('debug'):
-                        print(f"we can jump to {target} with model {ex.solver.model()}")
+                        print(f'we can jump to {target} with model {ex.solver.model()}')
                     new_ex = self.create_branch(ex, str(target_reachable), target)
                     stack.append((new_ex, step_id))
                 ex.solver.pop()
@@ -959,6 +1069,8 @@ class SEVM:
             storage  = deepcopy(ex.storage),
             balance  = deepcopy(ex.balance),
             #
+            block    = deepcopy(ex.block),
+            #
             calldata = ex.calldata,
             callvalue= ex.callvalue,
             caller   = ex.caller,
@@ -969,6 +1081,7 @@ class SEVM:
             jumpis   = deepcopy(ex.jumpis),
             output   = deepcopy(ex.output),
             symbolic = ex.symbolic,
+            prank    = deepcopy(ex.prank),
             #
             solver   = new_solver,
             path     = new_path,
@@ -977,6 +1090,7 @@ class SEVM:
             cnts     = deepcopy(ex.cnts),
             sha3s    = deepcopy(ex.sha3s),
             storages = deepcopy(ex.storages),
+            balances = deepcopy(ex.balances),
             calls    = deepcopy(ex.calls),
             failed   = ex.failed,
             error    = ex.error,
@@ -1024,6 +1138,11 @@ class SEVM:
                     out.append(ex)
                     continue
 
+                elif opcode == EVM.INVALID:
+                    ex.output = None
+                    out.append(ex)
+                    continue
+
                 elif opcode == EVM.REVERT:
                     ex.output = ex.st.ret()
                     out.append(ex)
@@ -1137,9 +1256,6 @@ class SEVM:
                     ex.solver.add(Extract(255, 160, f_origin()) == BitVecVal(0, 96))
                 elif opcode == EVM.ADDRESS:
                     ex.st.push(ex.this)
-                elif opcode == EVM.COINBASE:
-                    ex.st.push(f_coinbase())
-                    ex.solver.add(Extract(255, 160, f_coinbase()) == BitVecVal(0, 96))
                 elif opcode == EVM.EXTCODESIZE:
                     address = ex.st.pop()
                     if address in ex.code:
@@ -1157,18 +1273,24 @@ class SEVM:
                     ex.st.push(f_gas(con(ex.cnt_gas())))
                 elif opcode == EVM.GASPRICE:
                     ex.st.push(f_gasprice())
-                elif opcode == EVM.TIMESTAMP:
-                    ex.st.push(f_timestamp())
-                elif opcode == EVM.NUMBER:
-                    ex.st.push(f_blocknumber())
-                elif opcode == EVM.DIFFICULTY:
-                    ex.st.push(f_difficulty())
-                elif opcode == EVM.GASLIMIT:
-                    ex.st.push(f_gaslimit())
 
+                elif opcode == EVM.BASEFEE:
+                    ex.st.push(ex.block.basefee)
                 elif opcode == EVM.CHAINID:
-                #   ex.st.push(f_chainid())
-                    ex.st.push(con(1)) # for ethereum
+                    ex.st.push(ex.block.chainid)
+                elif opcode == EVM.COINBASE:
+                    ex.st.push(ex.block.coinbase)
+                elif opcode == EVM.DIFFICULTY:
+                    ex.st.push(ex.block.difficulty)
+                elif opcode == EVM.GASLIMIT:
+                    ex.st.push(ex.block.gaslimit)
+                elif opcode == EVM.NUMBER:
+                    ex.st.push(ex.block.number)
+                elif opcode == EVM.TIMESTAMP:
+                    ex.st.push(ex.block.timestamp)
+
+                elif opcode == EVM.PC:
+                    ex.st.push(con(ex.pc))
 
                 elif opcode == EVM.BLOCKHASH:
                     ex.st.push(f_blockhash(ex.st.pop()))
@@ -1298,6 +1420,8 @@ class SEVM:
         storage,
         balance,
         #
+        block,
+        #
         calldata,
         callvalue,
         caller,
@@ -1308,6 +1432,7 @@ class SEVM:
     #   jumpis,
     #   output,
         symbolic,
+    #   prank,
         #
         solver,
     #   path,
@@ -1316,6 +1441,7 @@ class SEVM:
     #   cnts,
     #   sha3s,
     #   storages,
+    #   balances,
     #   calls,
     #   failed,
     #   error,
@@ -1326,6 +1452,8 @@ class SEVM:
             storage  = storage,
             balance  = balance,
             #
+            block    = block,
+            #
             calldata = calldata,
             callvalue= callvalue,
             caller   = caller,
@@ -1336,6 +1464,7 @@ class SEVM:
             jumpis   = {},
             output   = None,
             symbolic = symbolic,
+            prank    = Prank(),
             #
             solver   = solver,
             path     = [],
@@ -1343,7 +1472,8 @@ class SEVM:
             log      = [],
             cnts     = defaultdict(int),
             sha3s    = [],
-            storages = [],
+            storages = {},
+            balances = {},
             calls    = [],
             failed   = False,
             error    = '',
